@@ -1,27 +1,29 @@
 use strict;
 use warnings;
-package JSON::Schema::Draft201909; # git description: v0.002-24-g55b80c8
+package JSON::Schema::Draft201909; # git description: v0.003-13-g3d882cb
 # vim: set ts=8 sts=2 sw=2 tw=100 et :
 # ABSTRACT: Validate data against a schema
 # KEYWORDS: JSON Schema data validation structure specification
 
-our $VERSION = '0.003';
+our $VERSION = '0.004';
 
 no if "$]" >= 5.031009, feature => 'indirect';
-use feature 'current_sub';
 use JSON::MaybeXS 1.004001 'is_bool';
 use Syntax::Keyword::Try 0.11;
-use Carp 'croak';
+use Carp qw(croak carp);
 use List::Util 1.33 qw(any pairs);
 use Mojo::JSON::Pointer;
 use Mojo::URL;
 use Safe::Isa;
+use Path::Tiny;
+use File::ShareDir 'dist_dir';
 use Moo;
 use MooX::TypeTiny 0.002002;
 use MooX::HandlesVia;
-use Types::Standard 1.010002 qw(Bool Int HasMethods Enum InstanceOf HashRef Dict);
+use Types::Standard 1.010002 qw(Bool Int Str HasMethods Enum InstanceOf HashRef Dict);
 use JSON::Schema::Draft201909::Error;
 use JSON::Schema::Draft201909::Result;
+use JSON::Schema::Draft201909::Document;
 use namespace::clean;
 
 has output_format => (
@@ -43,53 +45,6 @@ has max_traversal_depth => (
   default => 50,
 );
 
-has _resource_index => (
-  is => 'bare',
-  isa => HashRef[Dict[
-      # see JSON::MaybeXS::is_bool
-      ref => InstanceOf[qw(JSON::XS::Boolean Cpanel::JSON::XS::Boolean JSON::PP::Boolean)]|HashRef,
-      canonical_uri => InstanceOf['Mojo::URL'],
-    ]],
-  handles_via => 'Hash',
-  handles => {
-    _add_resources => 'set',
-    _get_resource => 'get',
-    _remove_resource => 'delete',
-    _resource_index => 'elements',
-  },
-  lazy => 1,
-  default => sub { {} },
-);
-
-before _add_resources => sub {
-  my $self = shift;
-  foreach my $pair (pairs @_) {
-    my ($key, $value) = @$pair;
-    if (my $existing = $self->_get_resource($key)) {
-      croak 'a schema resource is already indexed with uri "'.$key.'"'
-        # we allow overwriting canonical_uri = '' to allow for ad hoc evaluation of
-        # schemas that lack all identifiers altogether
-        if ($key ne '' and $existing->{canonical_uri} ne '')
-          and $existing->{ref} != $value->{ref}
-            or $existing->{canonical_uri} ne $value->{canonical_uri};
-    }
-
-    my $fragment = $value->{canonical_uri}->fragment;
-    croak sprintf('canonical_uri cannot contain an empty fragment (%s)', $value->{canonical_uri})
-      if defined $fragment and $fragment eq '';
-
-    croak sprintf('canonical_uri cannot contain a plain-name fragment (%s)', $value->{canonical_uri})
-      if ($fragment // '') =~ m{^[^/]};
-  }
-};
-
-has _json_decoder => (
-  is => 'ro',
-  isa => HasMethods[qw(encode decode)],
-  lazy => 1,
-  default => sub { JSON::MaybeXS->new(allow_nonref => 1, utf8 => 1) },
-);
-
 sub evaluate_json_string {
   my ($self, $json_data, $schema) = @_;
   my $data;
@@ -116,7 +71,16 @@ sub evaluate_json_string {
 sub evaluate {
   my ($self, $data, $schema) = @_;
 
-  $self->_find_all_identifiers($schema);
+  # TODO: move to $self->add_schema($schema)
+  my $document = JSON::Schema::Draft201909::Document->new(
+    # TODO canonical_uri => $self->base_uri,
+    schema => $schema,
+  );
+
+  $self->_add_resources(
+    map +( $_->[0] => +{ %{$_->[1]}, document => $document } ),
+      $document->_resource_pairs
+  );
 
   my $state = {
     base_uri => Mojo::URL->new,                   # TODO: will be set by a global attribute
@@ -151,6 +115,8 @@ sub evaluate {
   );
 }
 
+######## NO PUBLIC INTERFACES FOLLOW THIS POINT ########
+
 sub _eval {
   my ($self, $data, $schema, $state) = @_;
 
@@ -169,7 +135,7 @@ sub _eval {
 
   foreach my $keyword (
     # CORE KEYWORDS
-    qw($schema $id $anchor $ref $recursiveRef $recursiveAnchor $vocabulary $comment $defs),
+    qw($schema $id $anchor $recursiveAnchor $ref $recursiveRef $vocabulary $comment $defs),
     # VALIDATOR KEYWORDS
     qw(type enum const
       multipleOf maximum exclusiveMaximum minimum exclusiveMinimum
@@ -180,6 +146,8 @@ sub _eval {
     qw(allOf anyOf oneOf not if dependentSchemas
       items unevaluatedItems contains
       properties patternProperties additionalProperties unevaluatedProperties propertyNames),
+    # DISCONTINUED KEYWORDS
+    qw(definitions dependencies),
   ) {
     next if not exists $schema->{$keyword};
 
@@ -237,25 +205,36 @@ sub _eval_keyword_anchor {
   return 1;
 }
 
-sub _eval_keyword_ref {
+sub _eval_keyword_recursiveAnchor {
   my ($self, $data, $schema, $state) = @_;
 
-  assert_keyword_type($state, $schema, 'string');
+  assert_keyword_type($state, $schema, 'boolean');
+  return 1 if not $schema->{'$recursiveAnchor'} or exists $state->{recursive_anchor_uri};
 
-  my $uri = Mojo::URL->new($schema->{'$ref'})->base($state->{base_uri})->to_abs;
+  # record the canonical location of the current position, to be used against future resolution
+  # of a $recursiveRef uri -- as if it was the current location when we encounter a $ref.
+  my $uri = $state->{canonical_schema_uri} ? $state->{canonical_schema_uri}->clone : Mojo::URL->new;
+  $uri->fragment(($uri->fragment//'').$state->{schema_path});
+
+  $state->{recursive_anchor_uri} = $uri;
+  return 1;
+}
+
+sub _fetch_and_eval_ref_uri {
+  my ($self, $data, $schema, $state, $uri) = @_;
 
   my $fragment = $uri->fragment // '';
   my ($subschema, $canonical_uri);
-  # TODO: this will get less ugly when we move to actual document objects
   if (not length($fragment) or $fragment =~ m{^/}) {
     my $base = $uri->clone->fragment(undef);
-    my $document = Mojo::JSON::Pointer->new(($self->_get_resource($base) // {})->{ref});
-    $subschema = $document->get($fragment);
-    $canonical_uri = $uri;
+    if (my $resource = $self->_get_or_load_resource($base)) {
+      $subschema = $resource->{document}->get($resource->{path}.$fragment);
+      $canonical_uri = $uri;
+    }
   }
   else {
     if (my $resource = $self->_get_resource($uri)) {
-      $subschema = $resource->{ref};
+      $subschema = $resource->{document}->get($resource->{path});
       $canonical_uri = $resource->{canonical_uri}->clone; # this is *not* the anchor-containing URI
     }
   }
@@ -264,12 +243,36 @@ sub _eval_keyword_ref {
 
   return $self->_eval($data, $subschema,
     +{ %$state,
-      traversed_schema_path => $state->{traversed_schema_path}.$state->{schema_path}.'/$ref',
+      traversed_schema_path => $state->{traversed_schema_path}.$state->{schema_path}.'/'.$state->{keyword},
       canonical_schema_uri => $canonical_uri, # note: not canonical yet until $id is processed
       schema_path => '',
     });
 }
 
+sub _eval_keyword_ref {
+  my ($self, $data, $schema, $state) = @_;
+
+  assert_keyword_type($state, $schema, 'string');
+
+  my $uri = Mojo::URL->new($schema->{'$ref'})->base($state->{base_uri})->to_abs;
+  return $self->_fetch_and_eval_ref_uri($data, $schema, $state, $uri);
+}
+
+sub _eval_keyword_recursiveRef {
+  my ($self, $data, $schema, $state) = @_;
+
+  assert_keyword_type($state, $schema, 'string');
+
+  my $uri = Mojo::URL->new($schema->{'$recursiveRef'})
+    ->base($state->{recursive_anchor_uri})->to_abs;
+
+  abort($state, 'cannot resolve a $recursiveRef with a non-empty fragment against a'
+      .' $recursiveAnchor location with a canonical URI containing a fragment')
+    if $schema->{'$recursiveRef'} ne '#' and $state->{recursive_anchor_uri}->fragment;
+
+  return $self->_fetch_and_eval_ref_uri($data, $schema, $state, $uri);
+}
+
 sub _eval_keyword_vocabulary {
   my ($self, $data, $schema, $state) = @_;
 
@@ -840,6 +843,16 @@ sub _eval_keyword_propertyNames {
   return E($state, 'not all property names are valid');
 }
 
+sub _eval_keyword_definitions {
+  carp 'no-longer-supported "definitions" keyword present: this should be rewritten as "$defs"';
+  return 1;
+}
+
+sub _eval_keyword_dependencies {
+  carp 'no-longer-supported "dependencies" keyword present: this should be rewritten as "dependentSchemas" or "dependentRequired"';
+  return 1;
+}
+
 sub _is_type {
   my (undef, $type, $value) = @_;
 
@@ -946,52 +959,103 @@ sub _is_elements_unique {
   return 1;
 }
 
-sub _traverse_for_identifiers {
-  my ($data, $canonical_uri) = @_;
-  my $uri_fragment = $canonical_uri->fragment // '';
-  my %identifiers;
-  if (ref $data eq 'ARRAY') {
-    return map
-      __SUB__->($data->[$_], $canonical_uri->clone->fragment($uri_fragment.'/'.$_)),
-      0 .. $#{$data};
-  }
-  elsif (ref $data eq 'HASH') {
-    if (exists $data->{'$id'} and _is_type(undef, 'string', $data->{'$id'})) {
-      $canonical_uri = Mojo::URL->new($data->{'$id'})->base($canonical_uri)->to_abs;
-      # this might not be a real $id... wait for it to be encountered at runtime before dying
-      if (not length $canonical_uri->fragment) {
-        $canonical_uri->fragment(undef);
-        $identifiers{$canonical_uri} = { ref => $data, canonical_uri => $canonical_uri };
-      };
+has _resource_index => (
+  is => 'bare',
+  isa => HashRef[Dict[
+      canonical_uri => InstanceOf['Mojo::URL'],
+      path => Str,
+      document => InstanceOf['JSON::Schema::Draft201909::Document'],
+    ]],
+  handles_via => 'Hash',
+  handles => {
+    _add_resources => 'set',
+    _get_resource => 'get',
+    _remove_resource => 'delete',
+    _resource_index => 'elements',
+    _resource_keys => 'keys',
+    _add_resources_unsafe => 'set',
+  },
+  lazy => 1,
+  default => sub { {} },
+);
+
+before _add_resources => sub {
+  my $self = shift;
+  foreach my $pair (sort { $a->[0] cmp $b->[0] } pairs @_) {
+    my ($key, $value) = @$pair;
+    if (my $existing = $self->_get_resource($key)) {
+      # we allow overwriting canonical_uri = '' to allow for ad hoc evaluation of
+      # schemas that lack all identifiers altogether; we drop *all* resources from that document
+      $self->_remove_resource(
+          grep $self->_get_resource($_)->{document} == $existing->{document}, $self->_resource_keys)
+        if $key eq '';
+
+      croak 'uri "'.$key.'" conflicts with an existing schema resource'
+        if ($key ne '' and $existing->{canonical_uri} ne '')
+          and $existing->{path} ne $value->{path}
+            or $existing->{canonical_uri} ne $value->{canonical_uri};
     }
-    if (exists $data->{'$anchor'} and _is_type(undef, 'string', $data->{'$anchor'})) {
-      # we cannot change the canonical uri, or we won't be able to properly identify
-      # paths within this resource
-      my $uri = Mojo::URL->new->base($canonical_uri)->to_abs->fragment($data->{'$anchor'});
-      $identifiers{$uri} = { ref => $data, canonical_uri => $canonical_uri };
+    elsif ($self->CACHED_METASCHEMAS->{$key}) {
+      croak 'uri "'.$key.'" conflicts with an existing meta-schema resource';
     }
 
-    return
-      %identifiers,
-      map __SUB__->($data->{$_}, $canonical_uri->clone->fragment($uri_fragment.'/'.$_)), keys %$data;
+    my $fragment = $value->{canonical_uri}->fragment;
+    croak sprintf('canonical_uri cannot contain an empty fragment (%s)', $value->{canonical_uri})
+      if defined $fragment and $fragment eq '';
+
+    croak sprintf('canonical_uri cannot contain a plain-name fragment (%s)', $value->{canonical_uri})
+      if ($fragment // '') =~ m{^[^/]};
+  }
+};
+
+use constant CACHED_METASCHEMAS => {
+  'https://json-schema.org/draft/2019-09/hyper-schema'        => '2019-09/hyper-schema.json',
+  'https://json-schema.org/draft/2019-09/links'               => '2019-09/links.json',
+  'https://json-schema.org/draft/2019-09/meta/applicator'     => '2019-09/meta/applicator.json',
+  'https://json-schema.org/draft/2019-09/meta/content'        => '2019-09/meta/content.json',
+  'https://json-schema.org/draft/2019-09/meta/core'           => '2019-09/meta/core.json',
+  'https://json-schema.org/draft/2019-09/meta/format'         => '2019-09/meta/format.json',
+  'https://json-schema.org/draft/2019-09/meta/hyper-schema'   => '2019-09/meta/hyper-schema.json',
+  'https://json-schema.org/draft/2019-09/meta/meta-data'      => '2019-09/meta/meta-data.json',
+  'https://json-schema.org/draft/2019-09/meta/validation'     => '2019-09/meta/validation.json',
+  'https://json-schema.org/draft/2019-09/output/hyper-schema' => '2019-09/output/hyper-schema.json',
+  'https://json-schema.org/draft/2019-09/output/schema'       => '2019-09/output/schema.json',
+  'https://json-schema.org/draft/2019-09/schema'              => '2019-09/schema.json',
+};
+
+# returns the same as _get_resource
+sub _get_or_load_resource {
+  my ($self, $uri) = @_;
+
+  my $resource = $self->_get_resource($uri);
+  return $resource if $resource;
+
+  if (my $local_filename = $self->CACHED_METASCHEMAS->{$uri}) {
+    my $file = path(dist_dir('JSON-Schema-Draft201909'), $local_filename);
+    my $schema = $self->_json_decoder->decode($file->slurp_raw);
+    my $document = JSON::Schema::Draft201909::Document->new(schema => $schema);
+
+    $self->_add_resources_unsafe(
+      map +( $_->[0] => +{ %{$_->[1]}, document => $document } ),
+        $document->_resource_pairs
+    );
+
+    return $self->_get_resource($uri);
   }
 
-  return ();
-}
+  # TODO:
+  # - load from network or disk
+  # - handle such resources with $anchor fragments
 
-# traverse a schema document, find all identifiers and add them to the resource index.
-# internal only and subject to change!
-sub _find_all_identifiers {
-  my ($self, $schema) = @_;
+  return;
+};
 
-  my $base_uri = Mojo::URL->new;  # TODO: $self->base_uri->clone
-  my %identifiers = _traverse_for_identifiers($schema, $base_uri);
-
-  $identifiers{''} = { ref => $schema, canonical_uri => $base_uri }
-    if not "$base_uri" and ref $schema eq 'HASH' and not exists $schema->{'$id'};
-
-  $self->_add_resources(%identifiers);
-}
+has _json_decoder => (
+  is => 'ro',
+  isa => HasMethods[qw(encode decode)],
+  lazy => 1,
+  default => sub { JSON::MaybeXS->new(allow_nonref => 1, utf8 => 1) },
+);
 
 # shorthand for creating error objects
 use namespace::clean 'E';
@@ -1046,7 +1110,7 @@ JSON::Schema::Draft201909 - Validate data against a schema
 
 =head1 VERSION
 
-version 0.003
+version 0.004
 
 =head1 SYNOPSIS
 
@@ -1112,9 +1176,9 @@ that respects the Draft 2019-09 meta-schema at L<https://json-schema.org/draft/2
 
 The result is a L<JSON::Schema::Draft201909::Result> object, which can also be used as a boolean.
 
-=head2 CAVEATS
+=head1 LIMITATIONS
 
-=head3 TYPES
+=head2 TYPES
 
 Perl is a more loosely-typed language than JSON. This module delves into a value's internal
 representation in an attempt to derive the true "intended" type of the value. However, if a value is
@@ -1127,7 +1191,7 @@ option.
 
 For more information, see L<Cpanel::JSON::XS/MAPPING>.
 
-=head2 LIMITATIONS
+=head2 SPECIFICATION COMPLIANCE
 
 Until version 1.000 is released, this implementation is not fully specification-compliant.
 
@@ -1199,6 +1263,14 @@ use of plain-name fragments with C<$anchor>
 
 =back
 
+=head1 SECURITY CONSIDERATIONS
+
+The C<pattern> and C<patternProperties> keywords evaluate regular expressions from the schema.
+No effort is taken (at this time) to sanitize the regular expressions for embedded code or
+potentially pathological constructs that may pose a security risk, either via denial of service
+or by allowing exposure to the internals of your application. B<DO NOT RUN SCHEMAS FROM UNTRUSTED
+SOURCES.>
+
 =head1 SEE ALSO
 
 =over 4
